/******************************************************************
    Copyright (C) 2014,2015,2017,2019-2021  Henrik Carlqvist

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
******************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>

#include "nls.h"

#define VERSION "3.2"

struct config
{
   unsigned int uiNumJobs;
   unsigned int uiNumCmds;
   unsigned int uiRetries;
   size_t       tBlockSize;
   unsigned char *pcMagicBytes;
   unsigned char *pcMagicValid;
   size_t tMagicLen;
   char **aszCmds;
};

struct workers
{
   char *pBuf;
   size_t tInBufSize;
   size_t tOutBufSize;
   pid_t pid;
   int pipefd[2];
   struct workers *pNext;
};

static int iNextCmdNo=-1;

static void *pGetBuf=NULL;
static size_t tGetBufSize=0;

static pid_t tMainPid=0;
static unsigned int uiWantedJobs=0;

static void sig_handler(int iSignal)
{
   if(tMainPid != getpid())
   {
      kill(tMainPid, iSignal);
   }
   else
   {
      if(iSignal == SIGUSR2)
      {
	 uiWantedJobs++;
      }
      if(iSignal == SIGUSR1)
      {
	 if(uiWantedJobs > 1)
	 {
	    uiWantedJobs--;
	 }
      }
   }
} /* sig_handler */

static void print_version(void)
{
   printf(_("splitjob version %s\n"), VERSION);
   printf(_("Written by Henrik Carlqvist (source and support at splitjob.sourceforge.net)\n\n"));
   printf(_("Copyright (C) 2014,2015,2017,2019-2021 Henrik Carlqvist\n"));
   printf(_(
   "This program is free software; you can redistribute it and/or modify\n"
   "it under the terms of the GNU General Public License as published by\n"
   "the Free Software Foundation; either version 2 of the License, or\n"
   "(at your option) any later version.\n\n"

   "This program is distributed in the hope that it will be useful,\n"
   "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
   "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
   "GNU General Public License for more details.\n\n"

   "You should have received a copy of the GNU General Public License\n"
   "along with this program; if not, write to the Free Software\n"
   "Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.\n\n"));
} /* print_version */


static void print_help(const char *szCommand)
{
   printf(_("Usage:\n\t%s [options] [commands]\n"), szCommand);
   printf(
      _("Reads from stdin, splits and sends to multiple parallel invocations\n"
	"of commands and concatenates their output to stdout.\nOptions:\n"));
   printf(
      _("  -j <n>    Set number of parallel jobs (default number of commands)\n"));
   printf(
      _("  -b <size> Set block size for each job (default 1 MB)\n"));
   printf(
      _("  -m <hex>  Set magic bytes to separate blocks (default none)\n"));
   printf(
      _("  -B        Use bzip2 magic bytes to separate blocks\n"));
   printf(
      _("  -X        Use xz magic bytes to separate blocks\n"));
   printf(
      _("  -L        Use lzip magic bytes to separate blocks\n"));
   printf(
      _("  -r <n>    Set number of retries for failed jobs (default 3)\n"));
   printf(
      _("  -h        Display this help and exit\n"));
   printf(
      _("  -v        Show program version and copyright\n"));
   printf(
      _("Examples:\n"));
   printf("%s %s -j 4 bzip2 < bigfile > bigfile.bz2\n",
	  _("Use multiple local cores: "),
	  szCommand);
   printf("%s %s \"ssh h1 gzip\" \"ssh h2 gzip\" < f > f.gz\n",
	  _("Use remote machines:      "),
	  szCommand);
   printf("%s %s -j 2 -b 10M gzip < file > file.gz\n",
	  _("Big block reduce overhead:"),
	  szCommand);
} /* print_help */

static ssize_t handle_short_read(int fd, char *buf, size_t count)
{
   size_t tOut = 0;
   size_t tRead;

   if(count <= 0)
      return 0;
   
   do
   {
      if((tRead = read(fd, buf, count)) > 0)
      {
         count -= tRead;
         buf += tRead;
         tOut += tRead;
      }      
   } while((tRead > 0) && (count > 0));
   return tOut;
} /* handle_short_read */

static ssize_t handle_short_write(int fd, const char *buf, size_t count)
{
   size_t tOut = 0;
   size_t tWrite;

   if(count <= 0)
      return 0;
   
   do
   {
      if((tWrite = write(fd, buf, count)) > 0)
      {
         count -= tWrite;
         buf += tWrite;
         tOut += tWrite;
      }      
   } while((tWrite > 0) && (count > 0));
   return tOut;
} /* handle_short_write */

static void convert_magic_byte(const char *pc2BytesIn,
			       unsigned char *pcByteOut,
			       unsigned char *pcValid)
{
   long lRes;
   char *pcEnd;
   char szIn[3] = {pc2BytesIn[0], pc2BytesIn[1], 0};

   lRes = strtol(szIn, &pcEnd, 16);
   *pcValid = (*pcEnd == 0);
   *pcByteOut = lRes;
}

static void use_magic(const char *szMagic, struct config *psCfg)
{
   int i;
   
   if(psCfg->pcMagicBytes)
      free(psCfg->pcMagicBytes);
   if(psCfg->pcMagicValid)
      free(psCfg->pcMagicValid);
   psCfg->tMagicLen = strlen(szMagic)/2;
   psCfg->pcMagicBytes=malloc(psCfg->tMagicLen);
   psCfg->pcMagicValid=malloc(psCfg->tMagicLen);
   if((!psCfg->pcMagicBytes)||(!psCfg->pcMagicValid))
   {
      fprintf(stderr, _("Magic allocation failed\n"));
      exit(1);
   }	 
   for(i=0; i<psCfg->tMagicLen; i++)
      convert_magic_byte(&szMagic[i*2],
			 &(psCfg->pcMagicBytes[i]),
			 &(psCfg->pcMagicValid[i]));
} /* use_magic */

static ssize_t get_next_block_initiated_by_magic(const struct config *psCfg,
						 struct workers *pWorker)
{
   static unsigned char *pcPrevMagic=NULL;
   static int bGotEOF=0;
   size_t i;
   int iIn;
   size_t tOut=0;

   if(bGotEOF)
   {
      free(pcPrevMagic);
      pcPrevMagic=NULL;
      return 0;
   }
   if(!pcPrevMagic)
   {
      pcPrevMagic = malloc(psCfg->tMagicLen);
      if(!pcPrevMagic)
      {
	 fprintf(stderr, _("Magic allocation failed\n"));
	 exit(1);
      }
      for(i=0; i<psCfg->tMagicLen; i++)
      {
	 iIn = getchar();
	 if((iIn==EOF) ||
	    (psCfg->pcMagicValid[i] && (psCfg->pcMagicBytes[i] != iIn)))
	 {
	    fprintf(
	       stderr,
	       _("Input does not start with requested magic bytes!\n"));
	    exit(1);
	 }
	 pcPrevMagic[i] = iIn;
      }
   }
   if(pWorker->tOutBufSize < psCfg->tMagicLen)
   {
      pWorker->pBuf = realloc(pWorker->pBuf, 20*psCfg->tMagicLen);
      if(!pWorker->pBuf)
      {
	 fprintf(
	    stderr,
	    _("Unable to allocate read buffer!\n"));
	 exit(1);
      }
      pWorker->tOutBufSize = 20*psCfg->tMagicLen;
   }
   tOut = psCfg->tMagicLen;
   memcpy(pWorker->pBuf, pcPrevMagic, tOut);
   i=0;
   while(i<psCfg->tMagicLen)
   {
      iIn = getchar();
      if(iIn==EOF)
      {
	 bGotEOF=1;
      }
      if(!bGotEOF &&
	 (!(psCfg->pcMagicValid[i]) || (psCfg->pcMagicBytes[i] == iIn)))
      {
	 pcPrevMagic[i++] = iIn;
      }
      else
      {
	 if((tOut + i) >= pWorker->tOutBufSize)
	 {
	    pWorker->pBuf = realloc(pWorker->pBuf, 2*(tOut+i));
	    if(!pWorker->pBuf)
	    {
	       fprintf(
		  stderr,
		  _("Unable to allocate read buffer!\n"));
	       exit(1);
	    }
	    pWorker->tOutBufSize = 2*(tOut+i);
	 }
	 if(i)
	 {
	    memcpy(&(pWorker->pBuf[tOut]), pcPrevMagic, i);
	    tOut += i;
	    i=0;
	 }
	 if(!bGotEOF)
	 {
            if(!(psCfg->pcMagicValid[i]) || (psCfg->pcMagicBytes[i] == iIn))
            {
               pcPrevMagic[i++] = iIn;
            }
            else
            {
               pWorker->pBuf[tOut++] = iIn;
            }
	 }
	 else
	 {
	    return tOut;
	 }
      }
   }
   return tOut;
} /* get_next_block_initiated_by_magic */

static void init_config(int argc, char **argv, struct config *psCfg)
{
   int i;
   int n=0;
   memset(psCfg, 0, sizeof(struct config));
   psCfg->tBlockSize = 1024*1024;
   psCfg->uiRetries = 3;
   psCfg->uiNumJobs = 0;
   psCfg->pcMagicBytes = NULL;
   psCfg->pcMagicValid = NULL;
   psCfg->tMagicLen = 0;
   /* Find number of jobs and number of commands */
   for(i=1; i<argc; i++)
      if((i+1<argc) && !strcmp(argv[i], "-j"))
         psCfg->uiNumJobs = atoi(argv[++i]);
      else if((i+1<argc) && !strcmp(argv[i], "-b"))
      {
         psCfg->tBlockSize = atol(argv[++i]);
         if(argv[i][strlen(argv[i])-1]=='k')
            psCfg->tBlockSize *= 1024;
         if(argv[i][strlen(argv[i])-1]=='M')
            psCfg->tBlockSize *= 1024*1024;
         if(argv[i][strlen(argv[i])-1]=='G')
            psCfg->tBlockSize *= 1024*1024*1024;
         if(psCfg->tBlockSize < 1)
            psCfg->tBlockSize = 1;
      }
      else if((i+1<argc) && !strcmp(argv[i], "-m"))
         use_magic(argv[++i], psCfg);
      else if((i+1<argc) && !strcmp(argv[i], "-B"))
         use_magic("425a68**314159265359", psCfg);
      else if((i+1<argc) && !strcmp(argv[i], "-X"))
         use_magic("fd377a585a00", psCfg);
      else if((i+1<argc) && !strcmp(argv[i], "-L"))
         use_magic("4c5a4950", psCfg);
      else if((i+1<argc) && !strcmp(argv[i], "-r"))
         psCfg->uiRetries = atoi(argv[++i]);
      else if(!strcmp(argv[i], "-v"))
      {
  	 print_version();
	 exit(0);
      }
      else if(!strcmp(argv[i], "-h"))
      {
  	 print_help(argv[0]);
	 exit(0);
      }
      else
         psCfg->uiNumCmds++;
   psCfg->aszCmds = malloc(psCfg->uiNumCmds * sizeof(char *));
   /* Populate commands */
   for(i=1; i<argc; i++)
      if(!strcmp(argv[i], "-j"))
         ++i;
      else if(!strcmp(argv[i], "-b"))
         ++i;
      else if(!strcmp(argv[i], "-m"))
         ++i;
      else if(!strcmp(argv[i], "-B"))
         ;
      else if(!strcmp(argv[i], "-X"))
         ;
      else if(!strcmp(argv[i], "-L"))
         ;
      else if(!strcmp(argv[i], "-r"))
         ++i;
      else
         psCfg->aszCmds[n++]=argv[i];
   if(!psCfg->uiNumJobs)
      psCfg->uiNumJobs=psCfg->uiNumCmds;
   if(!psCfg->uiNumJobs)
      psCfg->uiNumJobs=1;
} /* init_config */

static void destroy_config(struct config *psCfg)
{
   if(psCfg)
      free(psCfg->aszCmds);
} /* destroy_config */

static const char *get_next_command(struct config scCfg)
{
   if(scCfg.uiNumCmds < 1)
      return "cat -";

   iNextCmdNo++;
   if(iNextCmdNo >= scCfg.uiNumCmds)
      iNextCmdNo = 0;
   return scCfg.aszCmds[iNextCmdNo];
} /* get_next_command */

static const char *get_prev_command(struct config scCfg)
{
   if(scCfg.uiNumCmds < 1)
      return "cat -";

   if(iNextCmdNo<0)
      iNextCmdNo=0;

   return scCfg.aszCmds[iNextCmdNo];
} /* get_prev_command */

static void destroy_workers(struct workers *pWorkers)
{
   struct workers *pNextWorker;

   while(pWorkers)
   {
      pNextWorker = pWorkers->pNext;
      if(pWorkers->pBuf)
      {
         free(pWorkers->pBuf);
	 pWorkers->pBuf = NULL;
      }
      free(pWorkers);
      pWorkers=pNextWorker;
   }
} /* destroy_workers */

static int add_one_worker(struct workers **ppAllWorkers,
			  struct workers **ppCurrentWorker, size_t tSize)
{
   struct workers *pNewWorker = calloc(1, sizeof(struct workers));
   if(!pNewWorker)
      return 0;
   pNewWorker->pBuf = malloc(tSize);
   if(!pNewWorker->pBuf)
   {
      free(pNewWorker);
      return 0;
   }
   pNewWorker->tInBufSize = 0;
   pNewWorker->tOutBufSize = tSize;
   pNewWorker->pNext = *ppCurrentWorker;
   if(*ppAllWorkers == *ppCurrentWorker)
   {
      *ppAllWorkers = pNewWorker;
   }
   else
   {
      struct workers *pBefore = *ppAllWorkers;
      while(pBefore && (pBefore->pNext != *ppCurrentWorker))
	 pBefore = pBefore->pNext;
      if(!pBefore) /* This should not happen */
      {
	 free(pNewWorker->pBuf);
	 free(pNewWorker);
	 return 0;
      }
      pBefore->pNext = pNewWorker;
   }
   *ppCurrentWorker = pNewWorker;
   return 1;
} /* add_one_worker */

static int remove_one_worker(struct workers **ppAllWorkers,
			     struct workers **ppCurrentWorker)
{
   struct workers *pNext = (**ppCurrentWorker).pNext;
   
   if(*ppAllWorkers == *ppCurrentWorker)
   {
      *ppAllWorkers = pNext;
   }
   else
   {
      struct workers *pBefore = *ppAllWorkers;
      while(pBefore && (pBefore->pNext != *ppCurrentWorker))
	 pBefore = pBefore->pNext;
      if(!pBefore) /* This should not happen */
	 return 0;
      pBefore->pNext = pNext;
   }
   (**ppCurrentWorker).pNext = NULL;
   destroy_workers(*ppCurrentWorker);
   if(pNext)
      *ppCurrentWorker = pNext;
   else
      *ppCurrentWorker = *ppAllWorkers;
   return 1;
} /* remove_one_worker */

static void destroy_all_but_one_workers(struct workers *pWorkers,
                                        struct workers *pKeepThis)
{
   struct workers *pNextWorker;

   while(pWorkers)
   {
      pNextWorker = pWorkers->pNext;
      pWorkers->pNext = NULL;
      if(pWorkers != pKeepThis)
	 destroy_workers(pWorkers);
      pWorkers=pNextWorker;
   }
} /* destroy_all_but_one_workers */

/* A new attempt might be made if the previous attempt failed */
static int restart_worker(struct workers *pAllWorkers,
                          struct workers *pWorker,
                          const char *szCmd)
{
   FILE *fpCmd;
   int readbufpipefd[2];
   pid_t tReadBufPid;
   size_t tReadTot=0;

   if(pipe(pWorker->pipefd))
   {
      fprintf(stderr, _("pipe failed\n"));
      return 0;
   }
   pWorker->pid = fork();
   if(pWorker->pid < 0)
   {
      fprintf(stderr, _("fork failed\n"));
      close(pWorker->pipefd[0]);
      close(pWorker->pipefd[1]);
      return 0;
   }
   if(pWorker->pid > 0)
   {
      close(pWorker->pipefd[1]);
      return 1;
   }
   /* We are now a child process, our parent will get our return value
      from the wait call */
   fclose(stdin); /* Needed not to mess with parent stdin FILE* */
   close(pWorker->pipefd[0]);
   /* Create a new pipe between command and reading buffer process */
   if(pipe(readbufpipefd))
      exit(-1);
   /* Save some RAM */
   destroy_all_but_one_workers(pAllWorkers, pWorker);
   free(pGetBuf);
   /* Create a reading buffer process */
   tReadBufPid = fork();
   if(tReadBufPid < 0)
      exit(-1);
   if(tReadBufPid > 0)
   {
      close(pWorker->pipefd[1]);
      if(dup2(readbufpipefd[1], STDOUT_FILENO) < 0)
	 exit(-1);
      close(readbufpipefd[1]);
      
      fpCmd = popen(szCmd, "w");
      if(!fpCmd)
	 exit(-1);
      if(fwrite(pWorker->pBuf, pWorker->tInBufSize, 1, fpCmd) != 1)
	 exit(-1);
      exit(WEXITSTATUS(pclose(fpCmd)));
   }
   /* We are now the reading buffer process */
   close(readbufpipefd[1]);
   do
   {
      tReadTot = handle_short_read(readbufpipefd[0], pWorker->pBuf,
				   pWorker->tOutBufSize);
      handle_short_write(pWorker->pipefd[1], pWorker->pBuf, tReadTot);
   }while(tReadTot);
   exit(0);
} /* restart_worker */

/* Returns 1 if a work started successfully, 0 at failure */
static int start_worker(struct workers *pAllWorkers,
                        struct workers *pWorker,
                        const char *szCmd, size_t tSize)
{
   pWorker->tInBufSize = tSize;
   return restart_worker(pAllWorkers, pWorker, szCmd);
} /* start_worker */

/* returns 1 at success, 0 at worker failure and -1 if worker wasn't started */
static int get_data_from_worker(struct workers *pWorker,
                                const struct config *psCfg,
                                int bTrashIt)
{
   size_t tReadSize=0;
   int bGotMoreData=0;
   int bFail=0;
   int iPidStatus;

   if(!pWorker->pid)
      return -1;
   if(!pGetBuf)
   {
      tGetBufSize=3*pWorker->tInBufSize;
      if(tGetBufSize < 2*psCfg->tBlockSize)
         tGetBufSize = 2*psCfg->tBlockSize;
   }
   while(!pGetBuf)
   {
      pGetBuf = malloc(tGetBufSize);
      if(!pGetBuf)
         tGetBufSize /= 2;
      if(!tGetBufSize)
         return 0;
   }
   do
   {
      if(tReadSize > 0)
      {
         bGotMoreData = 1;
         if(!bTrashIt)
            handle_short_write(STDOUT_FILENO, pGetBuf, tReadSize);
      }
      tReadSize = handle_short_read(pWorker->pipefd[0], pGetBuf, tGetBufSize);
   } while( tReadSize == tGetBufSize );
   close(pWorker->pipefd[0]);
   if(waitpid(pWorker->pid, &iPidStatus, 0) == -1)
   {
      if(bGotMoreData)
	 fprintf(stderr, _("Waitpid failed\n"));
      bFail = 1;
   }
   if(!WIFEXITED(iPidStatus))
   {
      if(bGotMoreData)
	 fprintf(stderr, _("Subtask did not exit\n"));
      bFail = 1;
   }
   else if(WEXITSTATUS(iPidStatus))
   {
      if(bGotMoreData)
	 fprintf(stderr, _("Subtask exited with error\n"));
      bFail = 1;
   }
   pWorker->pid = 0;
   if(bFail && bGotMoreData)
   {
      fprintf(stderr, _("Got too much data and failed!\n"));
      exit(1);
   }
   if(bFail)
      return 0;
   if(tReadSize > 0)
   {
      if(!bTrashIt)
         handle_short_write(STDOUT_FILENO, pGetBuf, tReadSize);
   }

   return 1;
} /* get_data_from_worker */

static unsigned int try_merge_with_next_job(struct workers **ppAllWorkers,
                                            struct workers **ppCurrentWorker,
                                            const struct config *psCfg,
                                            unsigned int uiJobsLeft)
{
   if(uiJobsLeft > 1)
   {
      struct workers *pNextWorker = (*ppCurrentWorker)->pNext;
      struct workers *pWorker = *ppCurrentWorker;
      
      if(!pNextWorker)
         pNextWorker = *ppAllWorkers;
      if(pNextWorker != pWorker)
      {
         /* consume any output data if next worker was started */
         get_data_from_worker(pNextWorker, psCfg, 1);
         
         if(pWorker->tOutBufSize <
            (pWorker->tInBufSize + pNextWorker->tInBufSize))
         {
            pWorker->pBuf = realloc(pWorker->pBuf,
                                    pWorker->tInBufSize +
                                    pNextWorker->tInBufSize);
            if(!pWorker->pBuf)
            {
               fprintf(
                  stderr,
                  _("Unable to allocate read buffer!\n"));
               exit(1);
            }
            pWorker->tOutBufSize =
               pWorker->tInBufSize + pNextWorker->tInBufSize;
         }
         memcpy(&(pWorker->pBuf[pWorker->tInBufSize]), pNextWorker->pBuf,
                pNextWorker->tInBufSize);
         fprintf(stderr, _("Merged sizes %ld + %ld = %ld\n"),
                 pWorker->tInBufSize, pNextWorker->tInBufSize,
                 pWorker->tInBufSize + pNextWorker->tInBufSize);
         pWorker->tInBufSize += pNextWorker->tInBufSize;
         if(remove_one_worker(ppAllWorkers, &pNextWorker))
         {
            uiJobsLeft--;
         }
      }
      else
         fprintf(stderr, _("Whoops, I seem to be the only job left!\n"));
   }
   else
   {
      if(uiWantedJobs > uiJobsLeft)
      {
         struct workers *pSavedWorker = *ppCurrentWorker;
         
         fprintf(stderr,
                 _("Out of running jobs, creating unstarted from stdin...\n"));
	 if(add_one_worker(ppAllWorkers, ppCurrentWorker, psCfg->tBlockSize))
         {
            *ppCurrentWorker = pSavedWorker;
	    uiJobsLeft++;
            (*ppAllWorkers)->tInBufSize =
               get_next_block_initiated_by_magic(psCfg, *ppAllWorkers);
            fprintf(stderr,
                    _("Got %ld bytes from stdin\n"),
                      (*ppAllWorkers)->tInBufSize);

            if((*ppAllWorkers)->tInBufSize)
               return try_merge_with_next_job(ppAllWorkers, ppCurrentWorker,
                                              psCfg, uiJobsLeft);
            else
            {
               fprintf(stderr, _("Got nothing more from stdin!\n"));
               remove_one_worker(ppAllWorkers, ppAllWorkers);
               uiJobsLeft--;
            }
         }
	 else
	 {
	    fprintf(stderr, _("Failed adding job, staying at %d jobs\n"),
		    uiJobsLeft);
         }
      }
      else
         fprintf(stderr, _("No more job to merge with!\n"));
   }
   return uiJobsLeft;
} /* try_merge_with_next_job */

int main(int argc, char **argv)
{
   struct config scCfg;
   struct workers *pAllWorkers, *pCurrentWorker;
   size_t tRead;
   int iRetries;
   int iResult;
   unsigned int uiJobsLeft;
   struct sigaction new_action;
   unsigned int uiPrevWantedJobs=0;

   nls_init();
   init_config(argc, argv, &scCfg);
   tMainPid = getpid();
   new_action.sa_handler=sig_handler;
   sigemptyset(&new_action.sa_mask);
   new_action.sa_flags=SA_RESTART;
   sigaction(SIGUSR1, &new_action, NULL);
   sigaction(SIGUSR2, &new_action, NULL); 

   uiJobsLeft = 0;
   uiWantedJobs = scCfg.uiNumJobs;
   uiPrevWantedJobs = uiWantedJobs;

   pAllWorkers = NULL;
   pCurrentWorker = pAllWorkers;

   do
   {
      iRetries = scCfg.uiRetries;
      if(uiWantedJobs > uiPrevWantedJobs)
      {
	 fprintf(stderr,
		 _("Increasing number of jobs to %d\n"), uiWantedJobs);
	 uiPrevWantedJobs = uiWantedJobs;
      }
      else if(uiWantedJobs < uiPrevWantedJobs)
      {
	 fprintf(stderr, _("Decreasing number of jobs to %d\n"),
		 uiWantedJobs);
	 uiPrevWantedJobs = uiWantedJobs;
      }
      if(uiWantedJobs > uiJobsLeft)
      {
	 if(add_one_worker(&pAllWorkers, &pCurrentWorker, scCfg.tBlockSize))
	    uiJobsLeft++;
	 else
	 {
	    fprintf(stderr, _("Failed adding job, staying at %d jobs\n"),
		    uiJobsLeft);
	    uiWantedJobs = uiJobsLeft;
	 }
	 
      }
      while(!get_data_from_worker(pCurrentWorker, &scCfg, 0) && (iRetries >0))
      {
	 fprintf(stderr, _("%s: Failed getting data from spawned job...\n"),
	    argv[0]);
         if(scCfg.tMagicLen && (iRetries>0))
         {
            /* Something might have gone wrong with position found by magic,
               maybe try to add data until next magic */
            fprintf(stderr, _("%s: Trying to merge data from next job...\n"),
                    argv[0]);
            uiJobsLeft = try_merge_with_next_job(&pAllWorkers,
                                                 &pCurrentWorker,
                                                 &scCfg,
                                                 uiJobsLeft);
            fprintf(stderr,
		    _("%s: Now only %d jobs and %d retries left...\n"),
                    argv[0], uiJobsLeft, iRetries);
         }
         while((iRetries-- >0)&&
	       (restart_worker(pAllWorkers, pCurrentWorker,
                               get_next_command(scCfg))!= 1))
	    fprintf(stderr, _("%s: Failed restarting '%s'\n"),
		    argv[0], get_prev_command(scCfg));
 	 if(iRetries >=0)
         {
	    fprintf(stderr, _("%s: Successfully restarted '%s'\n"
			      "but parallel advantage has been lost...\n"),
		    argv[0], get_prev_command(scCfg));
         }
               
      }
      if(iRetries <= 0) /* Failed getting expected data */
      {
	 fprintf(stderr, _("%s: Failed again, giving up!\n"),
		 argv[0]);
         return 2;
      }
      if(uiWantedJobs < uiJobsLeft)
      {
	 if(remove_one_worker(&pAllWorkers, &pCurrentWorker))
	 {
	    uiJobsLeft--;
	 }
	 else
	 {
	    fprintf(stderr,
		    _("Failed removing job, something is seriously wrong!\n"));
	    uiWantedJobs = uiJobsLeft;
	 }
	 tRead = 1;
      }
      else
      {
	 if(scCfg.tMagicLen)
	 {
	    tRead = get_next_block_initiated_by_magic(&scCfg, pCurrentWorker);
	 }
	 else
	 {
	    tRead = handle_short_read(STDIN_FILENO, pCurrentWorker->pBuf,
				      scCfg.tBlockSize);
	 }
	 if(tRead)
         {
	    while(!start_worker(pAllWorkers, pCurrentWorker,
				get_next_command(scCfg), tRead) &&
		  iRetries-- >0);
         }
	 if(iRetries < 0) /* Failed starting new job */
	 {
	    fprintf(stderr,
		    _("%s: Failed starting '%s' after %d retries, giving up!\n"),
		    argv[0], get_prev_command(scCfg), scCfg.uiRetries);
	    return 3;
	 }
	 pCurrentWorker = pCurrentWorker->pNext;
	 if(!pCurrentWorker)
	 {
	    pCurrentWorker = pAllWorkers;
	 }
      }
   } while(tRead > 0);

   /* At this point, all data is sent and some might be read. Now read the
      rest */
   while(uiJobsLeft)
   {
      iRetries = scCfg.uiRetries;
      while(!(iResult=get_data_from_worker(pCurrentWorker, &scCfg, 0)) &&
            (iRetries >0))
      {
	 fprintf(stderr, _("%s: Failed getting data from spawned job...\n"),
	    argv[0]);
         if(scCfg.tMagicLen && (iRetries>0))
         {
            /* Something might have gone wrong with position found by magic,
               maybe try to add data until next magic */
            fprintf(stderr, _("%s: Trying to merge data from next job...\n"),
                    argv[0]);
            uiJobsLeft = try_merge_with_next_job(&pAllWorkers,
                                                 &pCurrentWorker,
                                                 &scCfg,
                                                 uiJobsLeft);
            fprintf(stderr,
		    _("%s: Now only %d jobs and %d retries left...\n"),
                    argv[0], uiJobsLeft, iRetries);
         }
         while((iRetries-- >0) &&
	       (restart_worker(pAllWorkers,
                               pCurrentWorker,get_next_command(scCfg))!= 1))
	    fprintf(stderr, _("%s: Failed restarting '%s'\n"),
		    argv[0], get_prev_command(scCfg));
 	 if(iRetries >=0)
	    fprintf(stderr, _("%s: Successfully restarted '%s'\n"
			      "but parallel advantage has been lost...\n"),
		    argv[0], get_prev_command(scCfg));
      }
      if(iRetries <= 0) /* Failed getting expected data */
      {
	 fprintf(stderr, _("%s: Failed again, giving up!\n"),
		 argv[0]);
         return 2;
      }
      if(iResult == -1)
         uiJobsLeft--;
      pCurrentWorker = pCurrentWorker->pNext;
      if(!pCurrentWorker)
         pCurrentWorker = pAllWorkers;
   }
   destroy_workers(pAllWorkers);
   destroy_config(&scCfg);
   return 0;
} /* main */
